#!/usr/bin/env python
#coding:utf-8

##################################################
# MUBLAS Analyzer ver.1.4.28
##################################################

debug = 0

##############################################################
# header
##############################################################
import os
import sys
import csv
import string
import operator
import argparse
import numpy
import math
import subprocess
from datetime import datetime
from scipy import stats
import logging
from logging import getLogger,StreamHandler,DEBUG
logger = getLogger(__name__)
handler = logging.FileHandler('mublas_analyzer.log', 'a+')
handler.setLevel(DEBUG)
logger.setLevel(DEBUG)
logger.addHandler(handler)

G_gnuplot_script_header = "#gnuplot script\n" \
	+ 'set term postscript eps enhanced color\n' \
	+ 'set style line 11 lt 1 lw 1.2 ps 1.0 pt 6 lc rgb "#CC3300" # red\n' \
	+ 'set style line 21 lt 1 lw 1.2 ps 1.0 pt 2 lc rgb "#009900" # green\n' \
	+ 'set style line 31 lt 1 lw 1.2 ps 1.0 pt 8 lc rgb "#CC9900" # orange\n' \
	+ 'set style line 41 lt 1 lw 1.2 ps 1.0 pt 1 lc rgb "#3366CC" # blue\n' \
	+ 'set style line 81 lt 1 lw 1.2 ps 1.0 pt 7 lc rgb "#FF0099" # pink \n' \
	+ 'set style line 82 lt 2 lw 1.2 ps 1.0 pt 6 lc rgb "#FF0099" # pink \n' \
	+ 'set style line 91 lt 1 lw 1.2 ps 1.0 pt 7 lc rgb "#3399FF" # sky \n' \
	+ 'set style line 92 lt 2 lw 1.2 ps 1.0 pt 6 lc rgb "#3399FF" # sky \n' \
	+ 'set grid\n' \
	+ 'set size 0.5,0.6\n' \
	+ 'set xrange [0:]\n' \
	+ 'set yrange [0:]\n' \
	+ 'set key below\n'

G_dat_header = "%%M\tN\tK\tsec\tGFlops\tGB/s\titer\ttx\tty\tth\tyx\tsb\tbd\tbm\tWrpOcp\tBlkOcp\tGrdOcp\tcand\tkernel#"

##############################################################
# functions:
##############################################################
def ceiling (n, c):
	return math.ceil (n/c) * c

def import_csv_file (file_in):
	with open (file_in, 'r') as dat:
		csv_dat = list(csv.reader(dat, delimiter='\t'))
	for row in range(len(csv_dat)):
		for cell in range(len(csv_dat[row])):
			try:
				csv_dat[row][cell] = float(csv_dat[row][cell])
			except:
				pass
	return csv_dat

def export_csv_file (csv_dat, file_ou):
	for row in csv_dat:
		row = [str(text) for text in row]
		cmd = 'echo "' + '\t'.join(row) + '" >> ' + file_ou
		do (cmd)

def sort_csv_dat(csv_dat, key):
	header = csv_dat[0]
	body = sorted (csv_dat[1:], key=operator.itemgetter(header.index(key)), reverse=True)
	body.insert (0, header)
	return body

def list_to_str (list_dat):
	dat = ""
	for x in list_dat:
		dat += str(x) + "\t"
	return dat

def get_max_sub (csv_dat, key, rbase, ideal_perf):
	csv_sorted = sort_csv_dat (csv_dat, key)
	dat_perfm = [row[5] for row in csv_sorted]
	max_perfm = numpy.max(dat_perfm[1:])
	dat_occup = [row[csv_sorted[0].index(key)] for row in csv_sorted]
	max_perf_wo = []
	ifrom = 1
	irange = ceiling(dat_occup[1], rbase) 
	for i in range(1, len(dat_occup)):
		if (irange - dat_occup[i] > rbase or i == len(dat_perfm[1:])):
			ito = i
			if (i == len(dat_perfm[1:])):
				ito = i+1 
			if (ito >= ifrom):
				max_perfm_sub = numpy.max(dat_perfm[ifrom:ito])
				flag = 0
				if (max_perfm_sub >= max_perfm * ideal_perf):
					flag = 1
				max_perf_wo.append([round(irange-rbase,2),round(irange,2),round(max_perfm_sub,3),flag])
			ifrom = i
			irange = ceiling(dat_occup[ifrom], rbase)
	return max_perf_wo

def do (command):
	if (debug == 1):
		print "[cmd] " + command
	elif (debug == 2):
		print "[cmd] " + command
		os.system (command)
	else:
		os.system (command)

def debug_msg (ah, msg):
	logger.debug (datetime.now().strftime("%Y%m%d%H%M") + " < mublas_analyzer > " + msg)
	#print ("< mublas_analyzer > " + msg)

def analyze_limit (ah, dim_n, dat, key, rbase, ideal_perf, minmax):
	submat = get_max_sub (dat, key, rbase, ideal_perf)
	text = key + "\tGB/s\n"
	if (minmax == 'min'):
		limit = numpy.max(submat[0])
		for i in range(0,len(submat)):
			text += str(submat[i][0]) + "\t" + str(submat[i][1]) + "\t" + str(submat[i][2]) + "\t" + str(submat[i][3]) + "\n"
			if (submat[i][3]==1 and submat[i][0] < limit):
				limit = submat[i][0]
				if (i == len(submat)-1):
					limit = 0
	elif (minmax == 'max'):
		limit = numpy.min(submat[0])
		for i in range(0,len(submat)):
			text += str(submat[i][0]) + "\t" + str(submat[i][1]) + "\t" + str(submat[i][2]) + "\t" + str(submat[i][3]) + "\n"
			if (submat[i][3]==1 and submat[i][0] > limit):
				limit = submat[i][1]
				if (i == len(submat)-1):
					limit = 0
	file_ou = ah['file_base'] + '_n' + str(dim_n) + '_recipe_' + key + '.dat'
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = 'echo "' + text + '" > ' + ah['file_dir'] + file_ou
	do (cmd)
	return limit

def genplotscript (ah, param, x, y, color):
	text = "\n\n#" + param + "\n" \
		+ 'set title "' + param + '"\n' \
		+ 'set xlabel "' + param + '"\n'
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		# ---------------
		text += 'set output "fig_' + ah['file_base'] + '_n' + str(dim_n) + '_' + param + '.eps"\n' \
			+ 'plot \\\n' 
		filename = ah['file_base'] + "_n" + str(dim_n) + ".dat"
		text += '"' + filename + '" using ' + str(x) + ':' + str(y) + ' every 1 w points ls ' + str(color) + '\n'
		# ---------------
		dim_n += ah['dim_step']
	return text

def gen_analyzer_info (ah, filename):
	file = open (filename, 'w')
	print >> file, ah
	file.close()

# correlation coefficient
def do_gen_corr_file (ah):
	cmd = 'echo "%%N\ttx\tty\tth\tyx\tbd\tbm\two\tbo\tgo" > ' + ah['file_base'] + "_corr.dat"
	do (cmd)
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		# ---------------
		filename = ah['file_base'] + "_n" + str(dim_n) + ".dat"
		csv_dat = import_csv_file (filename)

		csv_dat.pop(0)
		dat_perf = [row[5] for row in csv_dat]
		dat_tx = [row[7] for row in csv_dat]
		dat_ty = [row[8] for row in csv_dat]
		dat_th = [row[9] for row in csv_dat]
		dat_yx = [row[10] for row in csv_dat]
		dat_bd = [row[12] for row in csv_dat]
		dat_bm = [row[13] for row in csv_dat]
		dat_wo = [row[14] for row in csv_dat]
		dat_bo = [row[15] for row in csv_dat]
		dat_go = [row[16] for row in csv_dat]

		cmd = 'echo "' \
			+ str(dim_n) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_tx)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_ty)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_th)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_yx)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_bd)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_bm)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_wo)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_bo)[0][1],3)) + '\t' \
			+ str(round(numpy.corrcoef (dat_perf, dat_go)[0][1],3)) + '\t' \
			+ '" >> ' + ah['file_base'] + '_corr.dat'
		do (cmd)
		dim_n += ah['dim_step']
		# ---------------

# add percentage information (actual/peak)
def do_add_percentage (ah):
	filename = ah['file_base'] + "_eval_mublas_offline.dat"
	csv_dat = import_csv_file (filename)
	dat_perf = [row[5] for row in csv_dat]
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		# ---------------
		dat_perf.pop(0)

		filename = ah['file_base'] + "_n" + str(dim_n) + ".dat"
		f = open (filename)
		lines = f.readlines ()
		f.close ()
		cmd = 'echo "' + G_dat_header + '\tper" > ' + filename
		do (cmd)
		for line in lines:
			if (line[0:1] != '%'):
				line = line.rstrip()
				line += "\t" + str(dat_perf[0])
				cmd = 'echo "' + line + '" >> ' + filename
				do (cmd)
		# ---------------
		dim_n += ah['dim_step']

# total number of candidates
def total_num_of_candidates (ah):
	count = 0
	i_tx = ah['tb_min']
	while (i_tx <= ah['tb_max']):
		i_ty_min = max (ah['tb_min'] / i_tx, 1)
		i_ty_max = max (ah['tb_max'] / i_tx, 1)
		i_ty = i_ty_min
		if (ah['tbdim'] == 1):
			i_ty = 1
			i_ty_max = 1
		while (i_ty <= i_ty_max):
			count += 1
			i_ty += 1
		i_tx += ah['tb_step']
	ah['totalnumber'] = count

def testing_exec (blas, ah, file_ou):
	cmd = "./testing_" + blas + "_" + ah['routine'] + " --nodisp=1 --mode=p" \
		+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
		+ " --range=" + str(ah['dim_start']) + ":" + str(ah['dim_stop']) + ":" + str(ah['dim_step']) \
		+ " > " + file_ou
	do (cmd)

##############################################################
# Evaluation
##############################################################
def do_evaluation (ah):
	debug_msg (ah, "do_evaluation: start.")
	file_io = ah['file_dir'] + ah['file_base'] + "_eval_mublas_online.dat"
	debug_msg (ah, 'creating "' + file_io + '"')
	testing_exec ("mublas", ah, file_io)

	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_cublas.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	testing_exec ("cublas", ah, file_ou)

	# open Tx&Ty file
	csv_dat = import_csv_file (file_io)
	dat_dim_n = [row[1] for row in csv_dat]
	dat_tx = [row[7] for row in csv_dat]
	dat_ty = [row[8] for row in csv_dat]
	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_mublas_online_static.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = "echo '%%' > " + file_ou
	do (cmd)
	for i in range(len(dat_dim_n)):
		cmd = "./testing_mublas_" + ah['routine'] + " --nodisp=1 --mode=p" \
		+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
		+ " --tx=" + str(dat_tx[i]) + " --ty=" + str(dat_ty[i]) \
		+ " --range=" + str(dat_dim_n[i]) + ":" + str(dat_dim_n[i]) + ":1 >> " \
		+ file_ou
		do (cmd)
	debug_msg (ah, "do_evaluation: done.")


##############################################################
# Execution
##############################################################
#DIMRANGE
def do_evaluation_for_all_tbsizes (ah):
	debug_msg (ah, "do_evaluation_for_all_tbsizes: start.")
	if (ah['dim_stop'] == 0 and ah['dim'] == 0):
		sys.exit()
	if (ah['dim_stop'] == 0):
		ah['dim_start'] = ah['dim']
		ah['dim_stop'] = ah['dim']
		ah['dim_step'] = 1

	# execution
	cmd = "mkdir -p " + ah['file_dir']
	do (cmd)

	file_ou = ah['file_dir'] + ah['file_base'] + "_info.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	gen_analyzer_info (ah, file_ou)
	total_num_of_candidates (ah)
	count = 0
	i_tx = ah['tb_min']
	while (i_tx <= ah['tb_max']):
		i_ty_min = max (ah['tb_min'] / i_tx, 1)
		i_ty_max = max (ah['tb_max'] / i_tx, 1)
		i_ty = i_ty_min
		if (ah['tbdim'] == 1):
			i_ty = 1
			i_ty_max = 1
		while (i_ty <= i_ty_max):
			count += 1
			file_ou = ah['file_dir'] + ah['file_base'] + "_tx" + str(i_tx) + "ty" + str(i_ty) + ".dat"
			debug_msg (ah, 'creating "' + file_ou + '" (' + str(count) + '/' + str(ah['totalnumber']) + ')')
			cmd = "./testing_mublas_" + ah['routine'] + " --nodisp=1 --mode=p" \
				+ " --devno=" + str(ah['devno']) + " --uplo=" + ah['uplo'] + " --trans=" + ah['trans'] + " --diag=" + ah['diag'] \
				+ " --range=" + str(ah['dim_start']) + ":" + str(ah['dim_stop']) + ":" + str(ah['dim_step']) \
				+ " --tx=" + str(i_tx) + " --ty=" + str(i_ty) \
				+ " > " + file_ou
			do (cmd)
			i_ty += 1
		i_tx += ah['tb_step']
	debug_msg (ah, "do_evaluation_for_all_tbsizes: done.")

##############################################################
# Analyze
##############################################################
#DIMRANGE

#in:txty_file(s), out:ndim_file
def do_gen_ndim_file (ah):
	debug_msg (ah, "do_gen_ndim_file: start.")
	count = 0
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		file_io = ah['file_dir'] + ah['file_base'] + "_n" + str(dim_n) + ".dat"
		debug_msg (ah, 'creating "' + file_io + '"')
		# gen dat files
		count += 1
		cmd = 'echo "' + G_dat_header + '" > ' + file_io
		do (cmd)
		i_tx = ah['tb_min']
		while (i_tx <= ah['tb_max']):
			i_ty_min = max (ah['tb_min'] / i_tx, 1)
			i_ty_max = max (ah['tb_max'] / i_tx, 1)
			i_ty = i_ty_min
			if (ah['tbdim'] == 1):
				i_ty = 1
				i_ty_max = 1
			while (i_ty <= i_ty_max):
				file_in = ah['file_dir'] + ah['file_base'] + "_tx" + str(i_tx) + "ty" + str(i_ty) + ".dat"
				cmd = "sed -n " + str(count) + "p " + file_in
				if (subprocess.check_output(cmd.strip().split(' ')).find('error') == -1):
					cmd = "sed -n " + str(count) + "p " + file_in + " >> " + file_io
					do (cmd)
				i_ty += 1
			i_tx += ah['tb_step']

		dim_n += ah['dim_step']
	debug_msg (ah, "do_gen_ndim_file: done.")

#DIMRANGE
def gen__statistics (ah):
	debug_msg (ah, "gen__statistics: start.")
	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_statistics.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = 'echo "%%N\tQ1\tmin\tmax\tQ3\tmean\tmedian\tcand" > ' + file_ou
	do (cmd)
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		file_in = ah['file_dir'] + ah['file_base'] + "_n" + str(dim_n) + ".dat"
		csv_dat = import_csv_file (file_in)
		dat_perf = [row[5] for row in csv_dat]
		dat_perf.pop(0)
		cmd = 'echo "' \
			+ str(dim_n) + '\t' \
			+ str(stats.scoreatpercentile(dat_perf, 25)) + '\t' \
			+ str(numpy.min(dat_perf)) + '\t' \
			+ str(numpy.max(dat_perf)) + '\t' \
			+ str(stats.scoreatpercentile(dat_perf, 75)) + '\t' \
			+ str(numpy.mean(dat_perf)) + '\t' \
			+ str(numpy.median(dat_perf)) + '\t' \
			+ str(ah['totalnumber']) \
			+ '" >> ' + file_ou
		do (cmd)
		dim_n += ah['dim_step']
	debug_msg (ah, "gen__statistics: done.")
	
#DIMRANGE
def gen__mublas_offline (ah):
	debug_msg (ah, "gen__mublas_offline: start.")
	file_ou = ah['file_dir'] + ah['file_base'] + "_eval_mublas_offline.dat"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = 'echo "' + G_dat_header + '" > ' + file_ou
	do (cmd)
	dim_n = ah['dim_start']
	while (dim_n <= ah['dim_stop']):
		file_in = ah['file_dir'] + ah['file_base'] + "_n" + str(dim_n) + ".dat"
		csv_dat = import_csv_file (file_in)
		csv_sorted = sort_csv_dat (csv_dat, 'GB/s')
		cmd = 'echo "' + list_to_str (csv_sorted[1]) + '" >> ' + file_ou
		do (cmd)
		dim_n += ah['dim_step']
	debug_msg (ah, "gen__mublas_offline: done.")


#DIM
def gen__recipe (ah):
	debug_msg (ah, "gen__recipe: start.")
	file_in = ah['file_dir'] + ah['file_base'] + "_n" + str(ah['dim']) + ".dat"
	file_ou = ah['file_dir'] + ah['file_base'] + '_n' + str(ah['dim']) + '_recipe.dat'
	debug_msg (ah, 'creating "' + file_ou + '"')

	csv_dat = import_csv_file (file_in)
	limit_tx = analyze_limit (ah, ah['dim'], csv_dat, 'tx',      8.0, ah['idealperf'], 'min')
	limit_ty = analyze_limit (ah, ah['dim'], csv_dat, 'ty',      1.0, ah['idealperf'], 'max')
	limit_th = analyze_limit (ah, ah['dim'], csv_dat, 'th',      8.0, ah['idealperf'], 'min')
	limit_yx = analyze_limit (ah, ah['dim'], csv_dat, 'yx',      0.1, ah['idealperf'], 'max')
	limit_bd = analyze_limit (ah, ah['dim'], csv_dat, 'bd',      8.0, ah['idealperf'], 'max')
	limit_bm = analyze_limit (ah, ah['dim'], csv_dat, 'bm',      0.1, ah['idealperf'], 'max') 
	limit_wo = analyze_limit (ah, ah['dim'], csv_dat, 'WrpOcp', 0.01, ah['idealperf'], 'min')
	limit_bo = analyze_limit (ah, ah['dim'], csv_dat, 'BlkOcp', 0.01, ah['idealperf'], 'min')
	limit_go = analyze_limit (ah, ah['dim'], csv_dat, 'GrdOcp', 0.01, ah['idealperf'], 'min')

	dat = ""
	dat += "idealperf\t" + str(ah['idealperf']) + "\n"
	dat += "\n"
	dat += "limit_tx\t" + str(limit_tx) + "\n"
	dat += "limit_ty\t" + str(limit_ty) + "\n"
	dat += "limit_th\t" + str(limit_th) + "\n"
	dat += "limit_yx\t" + str(limit_yx) + "\n"
	dat += "limit_bd\t" + str(limit_bd) + "\n"
	dat += "limit_bm\t" + str(limit_bm) + "\n"
	dat += "limit_wo\t" + str(limit_wo) + "\n"
	dat += "limit_bo\t" + str(limit_bo) + "\n"
	dat += "limit_go\t" + str(limit_go) + "\n"
	cmd = 'echo "' + dat + '" > ' + file_ou
	do (cmd)

	note = ah['device'] + "_" + str(ah['mublasversion']) + "_" + str(ah['idealperf']) + "_" + str(ah['date']) + "_dim=" + str(ah['dim'])
	body = "," + str(ah['cc']) + "," + ah['func_dname'].upper()[0:1] + "," + ah['func_dname'].upper()[1:] + ",\n"
	csv = "### note:  " + note + "\n"
	csv += "C,WrpOcpMin," + str(limit_wo) + body
	csv += "C,BlkOcpMin," + str(limit_bo) + body
	csv += "C,ThMin," + str(limit_th) + body
	csv += "C,TyPerTxMax," + str(limit_yx) + body

	file_ou = "../src/mublas_recipe.csv"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = 'echo "' + csv + '" >> ' + file_ou
	do (cmd)

	debug_msg (ah, "python mublas_gen_recipe.py")
	cmd = "cd ../src;python mublas_gen_recipe.py;cd ../analyzer"
	do (cmd)
	debug_msg (ah, "gen__recipe: done.")


def do_recompile (ah):
	debug_msg (ah, "do_recompile: start.")
	cmd = "cd ../;make clean --silent;make -j9 --silent;cd analyzer"
	do (cmd)

	cmd = "cp ../testing/testing_mublas_devinfo ."
	do (cmd)
	cmd = "cp ../testing/testing_mublas_" + ah['routine'] + " ."
	do (cmd)
	cmd = "cp ../testing/testing_cublas_" + ah['routine'] + " ."
	do (cmd)

	debug_msg (ah, "do_recompile: done.")

##############################################################
# Plot
##############################################################
def do_plot (ah):
	debug_msg (ah, "do_plot: start.")

	# script file (header)
	script = G_gnuplot_script_header  
	script += genplotscript (ah, 'wo', 15, 6, 11)
	script += genplotscript (ah, 'bo', 16, 6, 31)
	script += genplotscript (ah, 'go', 17, 6, 21)
	script += genplotscript (ah, 'tx', 8, 6, 41)
	script += genplotscript (ah, 'ty', 9, 6, 41)
	script += genplotscript (ah, 'th', 10, 6, 41)
	script += genplotscript (ah, 'yx', 11, 6, 41)

	# online vs offline
	plottitle = ah['func_dname'].upper() + ' on ' + ah['device']
	plottitle = plottitle.replace('_', '-')
	script = G_gnuplot_script_header
	script += "\n\n#stat\n" \
		+ 'set title "' + plottitle + '" offset 0,-0.5\n' \
		+ 'set xlabel "Problem Size"\n' \
		+ 'set ylabel "Performance [GB/s]" offset 1.5,0\n' \
		+ 'set xrange [0:' + str(ah['dim_stop']) + ']\n' \
		+ 'set yrange [0:*]\n' \
		+ 'set xtics ' +  str(ah['dim_stop']/4) + '\n' \
		+ 'set output "fig_' + ah['file_base'] + '_perf.eps"\n' \
		+ 'set boxwidth 0.4 relative\n' \
		+ 'set style fill empty\n' \
		+ 'plot \\\n' \
		+ '"' + ah['file_base'] + '_eval_statistics.dat" using 1:7:7:7:7 notitle w candlesticks ls 21,\\\n' \
		+ '"' + ah['file_base'] + '_eval_statistics.dat" using 1:2:3:4:5 title "All" w candlesticks whiskerbars ls 21,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_online.dat" using 2:6 title "Online" w linespoints ls 11,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_online_static.dat" using 2:6 title "Offline" w linespoints ls 31,\\\n' \
		+ '"' + ah['file_base'] + '_eval_cublas.dat" using 2:6 title "CUBLAS" w linespoints ls 41\n' \
		+ '\n#txty\n' \
		+ 'set output "fig_' + ah['file_base'] + '_txty.eps"\n' \
		+ 'set ylabel "# of threads"\n' \
		+ 'set autoscale y\n' \
		+ 'plot \\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_online.dat" using 2:8 title "Tx(online)" w linespoints ls 81,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_online.dat" using 2:9 title "Ty(online)" w linespoints ls 91,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_offline.dat" using 2:8 title "Tx(best)" w linespoints ls 82,\\\n' \
		+ '"' + ah['file_base'] + '_eval_mublas_offline.dat" using 2:9 title "Ty(best)" w linespoints ls 92\n'

	file_ou = ah['file_dir'] + ah['file_base'] + ".plt"
	debug_msg (ah, 'creating "' + file_ou + '"')
	cmd = "echo '" + script + "' > " + file_ou
	do (cmd)
	os.chdir (ah['file_dir'])
	debug_msg (ah, 'gnuplot mublas_analyzer_' + ah['func_dname'] + '.plt')
	cmd = "gnuplot " + ah['file_base'] + ".plt"
	do (cmd)
	os.chdir ("../../")
	debug_msg (ah, "do_plot: done.")
	
##############################################################
# main
##############################################################
def main ():
	argvs = sys.argv
	argc = len (argvs)
	parser = argparse.ArgumentParser (description="mublas analyzer")
	parser.add_argument ('--devno', dest='devno', action='store', default='0', help='devno') 
	parser.add_argument ('--routine', dest='routine', action='store', default='', help='routine') 
	parser.add_argument ('--uplo', dest='uplo', action='store', default='', help='uplo') 
	parser.add_argument ('--trans', dest='trans', action='store', default='', help='trans') 
	parser.add_argument ('--diag', dest='diag', action='store', default='', help='diag') 
	parser.add_argument ('--dim', dest='dim', action='store', default='0', help='dim') 
	parser.add_argument ('--dimrange', dest='dimrange', action='store', default='0:0:0', help='dimrange') 
	parser.add_argument ('--tbrange', dest='tbrange', action='store', default='0:0:0', help='tbrange') 
	parser.add_argument ('--tbdim', dest='tbdim', action='store', default='0', help='tbdim') 
	parser.add_argument ('--process', dest='process', action='store', default='', help='process') 
	parser.add_argument ('--idealperf', dest='idealperf', action='store', default='', help='idealperf') 
	args = parser.parse_args()

	option = ''
	routine = ''
	if (args.uplo != ''):
		option += args.uplo
	if (args.trans != ''):
		option += args.trans
	if (args.diag != ''):
		option += args.diag
	if (option != ''):
		routine += args.routine + "_" + option.lower()
	else:
		routine += args.routine 

	# get devinfo and create analyzer-handler
	ah = { 
		'devno': int(args.devno), \
		'routine': args.routine.lower(), \
		'uplo': args.uplo.lower(), \
		'trans': args.trans.lower(), \
		'diag': args.diag.lower(), \
		'tbdim': int(args.tbdim), \
		'tb_min': int(args.tbrange.split(':')[0]), \
		'tb_max': int(args.tbrange.split(':')[1]), \
		'tb_step': int(args.tbrange.split(':')[2]), \
		'dim': int(args.dim), \
		'dim_start': int(args.dimrange.split(':')[0]), \
		'dim_stop': int(args.dimrange.split(':')[1]), \
		'dim_step': int(args.dimrange.split(':')[2]), \
		'cc': '', \
		'device': '', \
		'idealperf': float(args.idealperf), \
		'mublasversion': '', \
		'date': datetime.now().strftime("%Y%m%d%H%M"), \
		'totalnumber' : 0, \
		'file_base': "mublas_analyzer_" + routine.lower(), \
		'file_dir': '', \
		'func_dname': routine.lower() \
	}

	debug_msg (ah, "---------------------------------------------" )

# ----------------------------
	do_recompile (ah)
# ----------------------------
	debug_msg (ah, "get device info...")
	cmd = "./testing_mublas_devinfo --devno=" + str(ah['devno']) + " > devinfo"
	do (cmd)
	f = open ("devinfo")
	lines = f.readlines ()
	f.close ()
	ah['cc'] = lines[0].rstrip()
	ah['device'] = lines[1].rstrip().replace(' ', '_')
	ah['mublasversion'] = lines[2].rstrip()
	ah['file_dir'] =  ah['device'] + "/" + routine.lower() + "/"

	if (ah['dim_stop'] == 0 and ah['dim'] == 0):
		debug_msg (ah, "dim error")
		sys.exit()
	if (ah['dim_stop'] == 0):
		debug_msg (ah, "ah[dim_stop] is empty, set ah[dim] to ah[dim_start] and ah[dim_stop]")
		ah['dim_start'] = ah['dim']
		ah['dim_stop'] = ah['dim']
		ah['dim_step'] = 1
	if (ah['dim'] == 0):
		debug_msg (ah, "ah[dim] is empty, set ah[dim_stop] to ah[dim]")
		ah['dim'] = ah['dim_stop']

	if (args.process == 'all'):
		do_evaluation_for_all_tbsizes (ah)
		do_gen_ndim_file (ah)
		gen__statistics (ah)
		gen__mublas_offline (ah)
		gen__recipe (ah)
		do_recompile (ah)
		do_evaluation (ah)
		do_plot (ah)
	elif (args.process == 'gen_recipe'):
		ah['dim_start'] = ah['dim_stop']
		do_evaluation_for_all_tbsizes (ah)
		do_gen_ndim_file (ah)
		gen__recipe (ah)
	elif (args.process == 'update_recipe'):
		gen__recipe (ah)
	elif (args.process == 'update_recipe2'):
		gen__recipe (ah)
		do_recompile (ah)
		do_evaluation (ah)
		do_plot (ah)
	elif (args.process == 'evaluation'):
		do_evaluation (ah)
		do_plot (ah)
	elif (args.process == 'replot'):
		do_plot (ah)

# ----------------------------

	cmd = "rm testing_mublas_devinfo devinfo"
	do (cmd)
	cmd = "rm testing_mublas_" + ah['routine'] 
	do (cmd)
	cmd = "rm testing_cublas_" + ah['routine'] 
	do (cmd)

	debug_msg (ah, "---------------------------------------------" )

##############################################################
# main
##############################################################
if __name__ == "__main__":
	main ()

