#include "mublas_devel.h"

/* ***********************************************
	Occupancy Calculation
   *********************************************** */   

__host__ __inline__ size_t ThSizeCheck (mublasHandle_t *mh) {
	if (mh->Tx * mh->Ty % mh->ThStep == 0) {
		if (mh->TxModTyZero)  {
			if (mh->Tx % mh->Ty == 0) 
				return 1;
			else
				return 0;
		} else {
			return 1;
		}
	} else {
		return 0;
	}
}

__host__ __inline__ void GetOcp (mublasHandle_t *mh) {
	const size_t ShmPerMpPmax = mh->devprop_sharedMemPerMultiprocessor;
	const size_t MpPerDev = mh->devprop_multiProcessorCount;
	const size_t RegPerMp = mh->devprop_regsPerMultiprocessor;
	const size_t RegPerTh = mh->funcatb.numRegs;
	const size_t RegPerWrp = ceilingEx ((float)(RegPerTh * WrpSize), mh->RegUnit); // Usage of registers per warp
	const size_t WrpPerMpRmax = min (floorEx ((float)RegPerMp / (float)RegPerWrp, mh->WrpGran), mh->WrpPerMpPmax); // Max warps per multiprocessor
	const size_t WrpPerBlk = (size_t) ceil ((float)(mh->Tx * mh->Ty) / (float)WrpSize); // Warps per block
	const size_t BlkPerMpRmax = (size_t) floor ((float)WrpPerMpRmax / (float)WrpPerBlk); // Blocks per multiprocessor (limited by register)
	const size_t BlkPerDev = mh->Bx * mh->By;
	size_t BlkPerMpSmax; // Blocks per multiprocessor (limited by shared memory)
	if (mh->ShmPerBlk == 0) {
		mh->BlkPerMpMax = min (BlkPerMpRmax, mh->BlkPerMpPmax); // Blocks per multiprocessor
	} else {
		BlkPerMpSmax = (size_t) floor ((float)ShmPerMpPmax / (float)mh->ShmPerBlk);
		mh->BlkPerMpMax = min (min (BlkPerMpRmax, BlkPerMpSmax), mh->BlkPerMpPmax); // Blocks per multiprocessor
	}
	// Block-Occupancy
	mh->BlkOcp = (float)mh->BlkPerMpMax / (float)mh->BlkPerMpPmax;
	const size_t BlkPerDevMax = mh->BlkPerMpMax * MpPerDev;
	// Grid-Occupancy
	mh->GrdOcp = (float)BlkPerDev / (float)ceilingEx ((float)BlkPerDev, BlkPerDevMax);
	size_t WrpPerMpMax = WrpPerBlk * mh->BlkPerMpMax; // Warps per multiprocessor
	// Warp-Occupancy
	mh->WrpOcp = (float)WrpPerMpMax / (float)mh->WrpPerMpPmax; // Warp Occupancy
}


__host__ void determine_best_thread_block_size_2d (mublasHandle_t *mh) {
	size_t Tx, TxOptimal, TxStart, TxStop, TxStep;
	size_t Ty, TyOptimal, TyStart, TyStop, TyStep, TyMin, TyMax;
	float GrdOcpOld = 0., BlkOcpOld = 0., WrpOcpOld = 0.;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;

	// --- initial values -------------------------------------
	if (mh->TxPrefer == 'L') {
		TxStart = mh->TxMin;
		TxStop  = mh->TxMax;
		TxStep  = mh->TxStep;
	} else {
		TxStart = mh->TxMax;
		TxStop  = mh->TxMin;
		TxStep  = -mh->TxStep;
	}
	mh->count = 0;
	if (mh->autotuning) printf ("\n");
	// --- Tx ------------------------------------------------
	Tx = TxStart;
	while (Tx) {
		mh->Tx = Tx;
		TyMin = (size_t)ceil (mh->ThMin / (float)mh->Tx);
		TyMax = max (1, (size_t)floor ((float)mh->ThMax / (float)mh->Tx));
		if (mh->TyPerTxMax != 0)
			TyMax = min (TyMax, mh->Tx * mh->TyPerTxMax);
		if (mh->TyPrefer == 'L') {
			TyStart = TyMin;
			TyStop  = TyMax;
			TyStep  = 1;
		} else {
			TyStart = TyMax;
			TyStop  = TyMin;
			TyStep  = -1;
		}
		// --- Ty -----------------------------------------------
		Ty = TyStart;
		while (Ty) {
			mh->Ty = Ty;
			if (ThSizeCheck(mh)) {
				mh->count++;
				getBlkPerGrd (mh);
				getShmPerBlk (mh);
				GetOcp (mh);
				if (mh->autotuning) {
					size_t dimGrdOcpMax = Tx * mh->N_T * mh->BlkPerMpMax * mh->devprop_multiProcessorCount;
					dimGrdOcpMax = ceilingEx (10240, dimGrdOcpMax);
					printf ("[AT]%d\t%d\t%d\t%1.3f\t%1.3f\t%1.3f\t%1.3f\t%d\n",mh->count,Tx, Ty, (float)Ty/(float)Tx, mh->WrpOcp, mh->BlkOcp, mh->GrdOcp, dimGrdOcpMax);
				}
				if ( mh->GrdOcp >= min (GrdOcpOld, mh->GrdOcpMin) &&
					 mh->BlkOcp >= min (BlkOcpOld, mh->BlkOcpMin) &&
					 mh->WrpOcp >= min (WrpOcpOld, mh->WrpOcpMin) ) {
					GrdOcpOld = mh->GrdOcp;
					BlkOcpOld = mh->BlkOcp;
					WrpOcpOld = mh->WrpOcp;
					TxOptimal = Tx;
					TyOptimal = Ty;
				}
			}
			if      (mh->TyPrefer == 'L' && Ty >= TyStop) break;
			else if (mh->TyPrefer == 'S' && Ty <= TyStop) break;
			Ty += TyStep;
		// --- Ty -----------------------------------------------
		}
		if      (mh->TxPrefer == 'L' && Tx >= TxStop) break;
		else if (mh->TxPrefer == 'S' && Tx <= TxStop) break;
		Tx += TxStep;
	}
	// --- Tx ------------------------------------------------
	mh->Tx = TxOptimal;
	mh->Ty = TyOptimal;
}
		
__host__ void determine_best_thread_block_size_1d (mublasHandle_t *mh) {
	size_t Tx, TxOptimal, TxStart, TxStop, TxStep;
	float GrdOcpOld = 0., BlkOcpOld = 0., WrpOcpOld = 0.;
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;
	// --- initial values -------------------------------------
	if (mh->TxPrefer == 'L') {
		TxStart = max (mh->ThMin, mh->TxMin);
		TxStop  = min (mh->ThMax, mh->TxMax);
		TxStep  = mh->TxStep;
	} else if (mh->TxPrefer == 'S') {
		TxStart = min (mh->ThMax, mh->TxMax);
		TxStop  = max (mh->ThMin, mh->TxMin);
		TxStep  = -mh->TxStep;
	} 
	mh->count = 0;
	mh->Ty = 1;
	if (mh->autotuning) printf ("\n");
	// --- Tx ------------------------------------------------
	Tx = TxStart;
	while (Tx) {
		mh->Tx = Tx;
		if (ThSizeCheck(mh)) {
			mh->count++;
			getBlkPerGrd (mh);
			getShmPerBlk (mh);
			GetOcp (mh);
			if (mh->autotuning) {
				size_t dimGrdOcpMax = Tx * mh->N_T * mh->BlkPerMpMax * mh->devprop_multiProcessorCount;
				dimGrdOcpMax = ceilingEx (10240, dimGrdOcpMax);
				printf ("[AT]%d\t%d\t%1.3f\t%1.3f\t%1.3f\t%d\n",mh->count,Tx, mh->WrpOcp, mh->BlkOcp, mh->GrdOcp, dimGrdOcpMax);
			}
			if ( mh->GrdOcp >= min (GrdOcpOld, mh->GrdOcpMin) &&
				 mh->BlkOcp >= min (BlkOcpOld, mh->BlkOcpMin) &&
				 mh->WrpOcp >= min (WrpOcpOld, mh->WrpOcpMin)) {
				GrdOcpOld = mh->GrdOcp;
				BlkOcpOld = mh->BlkOcp;
				WrpOcpOld = mh->WrpOcp;
				TxOptimal = Tx;
			}
		}
		if      (mh->TxPrefer == 'L' && Tx >= TxStop) break;
		else if (mh->TxPrefer == 'S' && Tx <= TxStop) break;
		Tx += TxStep;
	}
	// --- Tx ------------------------------------------------
	mh->Tx = TxOptimal;
	mh->Ty = 1;
}

// phase 2: determining optimal thread-block size
__host__ void mublas_optimizer (mublasHandle_t *mh) {
	void (*getShmPerBlk)(mublasHandle_t *mh);
	void (*getBlkPerGrd)(mublasHandle_t *mh);
	size_t (*getTxMax)(mublasHandle_t *mh);
	getShmPerBlk = mh->func_getShmPerBlk;
	getBlkPerGrd = mh->func_getBlkPerGrd;
	getTxMax     = mh->func_getTxMax;

	// --- Initial Tx & Ty -------------------------------
	// 1. TxStep (0=auto)
	if (mh->TxStep == 0) mh->TxStep = 128 / (mh->N_one * mh->N_T); // 128bytes/16bytes(vec)=8thread
	// 2. ThStep (0=auto)
	if (mh->ThStep == 0) mh->ThStep = mh->TxStep;//WrpSize;
	// 3. ThMin (0=auto)
	if (mh->ThMin == 0) mh->ThMin = mh->WrpGran * WrpSize;
	// 4. ThMax
	if (mh->ThMax == 0) mh->ThMax = mh->funcatb.maxThreadsPerBlock;
	// 5. TxMin
	if (mh->TxMin == 0) mh->TxMin = mh->TxStep;
	// 6. TxMax
	if (mh->TxMax == 0) mh->TxMax = min (mh->ThMax, getTxMax (mh));
	// 0. auto-tuning mode
	if (mh->autotuning) {
		mh->TxMin = mh->TxStep;
		mh->TxMax = mh->ThMax;
	}
	// ---------------------------------------------------

	if (mh->TxStatic == 0) { 
		if      (mh->BlkDim == 2) determine_best_thread_block_size_2d (mh);
		else if (mh->BlkDim == 1) determine_best_thread_block_size_1d (mh);
		else printf ("mublas error: unsupported BlkDim\n");
	} else { // Kernel is launched with specified Tx and Ty (static mode)
		mh->Tx = mh->TxStatic;
		mh->Ty = mh->TyStatic;
		if (mh->Tx * mh->Ty > mh->funcatb.maxThreadsPerBlock) {
			printf ("mublas error: too many resource requested\n");
			exit (1);
		}
	}
	getBlkPerGrd (mh);
	getShmPerBlk (mh);
	GetOcp (mh);
}

void
mublasRecipeDefault (mublasHandle_t *mh) {
	// --------------------------------------------------------
	// default recipe (for all devices and all routines)
	mh->WrpOcpMin = (float)mh->WrpGran / (float)mh->WrpPerMpPmax; 
	mh->BlkOcpMin = 2.0 / (float)mh->BlkPerMpPmax;
	mh->GrdOcpMin = 0.95;
	mh->TxMin = 0; //0=auto:=TxStep
	mh->TxMax = 0;
	mh->TxStep = 0; //0=auto:128/sizeof(val)
	mh->ThMin = 0; //0=auto:ThMin=TxStep
	mh->ThMax = 0; //0=auto:maxThreadsPerBlock
	mh->ThStep = 0; //0=auto=TxStep
	mh->TxPrefer = 'L';
	mh->TyPrefer = 'S';
	mh->TyPerTxMax = 4.0; //0=none
	mh->TxModTyZero = 0;
	// --------------------------------------------------------
}

/* ***********************************************
	mublas functions
   *********************************************** */   

void
mublasCreate (mublasHandle_t *mh) {
    int32_t driverVersion = 0;
	int32_t runtimeVersion = 0;
	int32_t devid;
	cudaDriverGetVersion(&driverVersion);
	cudaRuntimeGetVersion(&runtimeVersion);

	// Device Properties
	cudaDeviceProp devprop;  
	cudaGetDevice (&devid);
	cudaGetDeviceProperties (&devprop, devid);
	mh->devprop_multiProcessorCount = devprop.multiProcessorCount;
	mh->devprop_sharedMemPerMultiprocessor = devprop.sharedMemPerMultiprocessor;
	mh->devprop_regsPerMultiprocessor = devprop.regsPerMultiprocessor;
	mh->devprop_sharedMemPerBlock = devprop.sharedMemPerBlock;
	mh->devprop_regsPerBlock = devprop.regsPerBlock;
	mh->devprop_major = devprop.major;
	mh->cc = devprop.major * 100 + devprop.minor * 10;
	mh->driverVersion = driverVersion;
	mh->runtimeVersion = runtimeVersion;
	mh->streamId = 0;

	// work space allocation
	float *w;
	cudaMalloc ((void **) &w, sqrt((float)devprop.totalGlobalMem/(sizeof(double)*4))*sizeof(double)*4);
	mh->work_addr = w;
	float *w2;
	cudaMalloc ((void **) &w2, sqrt((float)devprop.totalGlobalMem/(sizeof(double)*4))*sizeof(double)*4);
	mh->work_addr2 = w2;

	// --------------------------------------------------------
	// init
	mh->mublasVersion = MUBLAS_VERSION;
	mh->mublaserror = 1111; // set 1111 when mublasCreate called
	mh->kernelnumber = 0;
	mh->count = 0;
	mh->debug = 0;
	mh->autotuning = 0;
	mh->WrpOcp = 0.;
	mh->BlkOcp = 0.;
	mh->GrdOcp = 0.;
	mh->Tx = 0;
	mh->Ty = 0;
	mh->Bx = 0;
	mh->By = 0;
	mh->flops = 0.;
	mh->flops_prev = 0.;
	mh->WrpOcp_prev = 0.;
	mh->BlkOcp_prev = 0.;
	// --------------------------------------------------------

	// default value (architecture dependent)
	switch (mh->cc) {
		case 200:
			mh->WrpPerMpPmax = 48;
			mh->BlkPerMpPmax = 8;
			mh->RegUnit = 64;
			mh->ShmUnit = 128;
			mh->WrpGran = 2;
			mh->Cores = 32;
			mh->DPCores = 16;
			break;
		case 210:
			mh->WrpPerMpPmax = 48;
			mh->BlkPerMpPmax = 8;
			mh->RegUnit = 64;
			mh->ShmUnit = 128;
			mh->WrpGran = 2;
			mh->Cores = 48;
			mh->DPCores = 6;
			break;
		case 300:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 320:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 350:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 370:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 16;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 192;
			mh->DPCores = 64;
			break;
		case 500:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 32;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 128;
			mh->DPCores = 4;
			break;
		case 520:
			mh->WrpPerMpPmax = 64;
			mh->BlkPerMpPmax = 32;
			mh->RegUnit = 256;
			mh->ShmUnit = 256;
			mh->WrpGran = 4;
			mh->Cores = 128;
			mh->DPCores = 4;
			break;
		default: // non-supported architecture
			mh->mublaserror = 9999;
			break;
	}
	mublasRecipeDefault (mh);
}

void
mublasSetStream (mublasHandle_t *mh, cudaStream_t stream) {
	mh->streamId = stream;
}

void
mublasDestroy (mublasHandle_t *mh) {
	cudaFree (mh->work_addr);
	cudaFree (mh->work_addr2);
}

/* ***********************************************
	mublas functions (multi-GPU support)
   *********************************************** */   

void mublasSetMatrixMgpu1D (size_t m, size_t n, void* hstAddr, size_t ld_hst, void** devAddr, size_t ld_dev, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	char* devAddr_;
	char* hstAddr_ = (char*)hstAddr;
	size_t nr, hst_offset, dev_offset;
	dev_offset = 0;
	hst_offset = 0;

	nr = n;
	igpu = 0;
	while (nr > bs) {
		cudaSetDevice (igpu%ngpus);
		devAddr_ = (char*)devAddr[igpu%ngpus];
		cublasSetMatrixAsync (m, bs, word_size, hstAddr_+hst_offset, ld_hst, devAddr_+dev_offset, ld_dev, 0);
		// update
		nr -= bs;
		igpu++;
		hst_offset += ld_hst * bs * word_size;
		if (igpu%ngpus == 0) dev_offset += ld_dev * bs * word_size;
	} 
	cudaSetDevice (igpu%ngpus);
	devAddr_ = (char*)devAddr[igpu%ngpus];
	cublasSetMatrixAsync (m, nr, word_size, hstAddr_+hst_offset, ld_hst, devAddr_+dev_offset, ld_dev, 0);

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

void mublasSetVectorMgpu1D (size_t n, void* hstAddr, void** devAddr, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	char* devAddr_;
	char* hstAddr_ = (char*)hstAddr;
	size_t nr, hst_offset, dev_offset;
	dev_offset = 0;
	hst_offset = 0;

	nr = n;
	igpu = 0;
	while (nr > bs) {
		cudaSetDevice (igpu%ngpus);
		devAddr_ = (char*)devAddr[igpu%ngpus];
		cublasSetVectorAsync (bs, word_size, hstAddr_+hst_offset, 1, devAddr_+dev_offset, 1, 0);
		// update
		nr -= bs;
		igpu++;
		hst_offset += bs * word_size;
		if (igpu%ngpus == 0) dev_offset += bs * word_size;
	} 
	cudaSetDevice (igpu%ngpus);
	devAddr_ = (char*)devAddr[igpu%ngpus];
	cublasSetVectorAsync (nr, word_size, hstAddr_+hst_offset, 1, devAddr_+dev_offset, 1, 0);

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

void mublasGetVectorMgpu1D (size_t n, void** devAddr, void* hstAddr, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	char* devAddr_;
	char* hstAddr_ = (char*)hstAddr;
	size_t nr, hst_offset, dev_offset;
	dev_offset = 0;
	hst_offset = 0;

	nr = n;
	igpu = 0;
	while (nr > bs) {
		cudaSetDevice (igpu%ngpus);
		devAddr_ = (char*)devAddr[igpu%ngpus];
		cublasGetVectorAsync (bs, word_size, devAddr_+dev_offset, 1, hstAddr_+hst_offset, 1, 0);
		// update
		nr -= bs;
		igpu++;
		hst_offset += bs * word_size;
		if (igpu%ngpus == 0) dev_offset += bs * word_size;
	} 
	cudaSetDevice (igpu%ngpus);
	devAddr_ = (char*)devAddr[igpu%ngpus];
	cublasGetVectorAsync (nr, word_size, devAddr_+dev_offset, 1, hstAddr_+hst_offset, 1, 0);

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

void mublasMallocMatrixMgpu1D (size_t m, size_t n, void** devAddr, size_t ld_dev, uint32_t ngpus, size_t bs, size_t word_size) {
	uint32_t igpu;
	size_t pitch, nr, ni[NGPU_MAX] = {0};

	nr = n;
	igpu = ngpus;
	while (nr >= bs) {
		ni[igpu%ngpus] += bs;
		nr -= bs;
		igpu++;
	} 
	ni[igpu%ngpus] += nr;

	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaMalloc ((void **) &devAddr[igpu], word_size * ld_dev * ni[igpu]);
	}
	
	for (igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize ();
	}
}

/* ***********************************************
	mublas functions (multi-GPU support, 1D-block)
   *********************************************** */   

void mublasMallocVectorMgpu (mublasMgpuVectorHandle_t* mmah, size_t n, size_t word_size, size_t ngpus) {
	size_t i;
	mmah->word_size = word_size;
	mmah->ngpus = ngpus;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		mmah->n[i] = n;
		cudaMalloc ((void **) &mmah->devAddr[i], mmah->word_size * mmah->n[i]);
	}
}

void mublasMallocVectorMgpuDistributed (mublasMgpuVectorHandle_t* mmah, size_t n, size_t word_size, size_t ngpus) {
	size_t i;
	mmah->word_size = word_size;
	mmah->ngpus = ngpus;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		mmah->n[i] = (i == ngpus-1) ? n-ceil(n/ngpus)*(ngpus-1) : ceil(n/ngpus);
		cudaMalloc ((void **) &mmah->devAddr[i], mmah->word_size * mmah->n[i]);
	}
}

void mublasMallocMatrixMgpuDistributed (mublasMgpuMatrixHandle_t* mmah, char trans, size_t m, size_t n, size_t word_size, size_t ngpus) {
	size_t i;
	size_t pitch;
	mmah->word_size = word_size;
	mmah->ngpus = ngpus;
	if (trans == 'N' || trans == 'n') {
		for (i = 0; i < mmah->ngpus; i++) {
			mmah->m[i] = (i == ngpus-1) ? m-ceil(m/ngpus)*(ngpus-1) : ceil(m/ngpus);
			mmah->n[i] = n;
			cudaSetDevice (i);
			cudaMallocPitch ((void **) &mmah->devAddr[i], &pitch, mmah->word_size * mmah->m[i], mmah->n[i]);
			mmah->ld[i] = pitch/mmah->word_size;
		}
	} else {
		for (i = 0; i < mmah->ngpus; i++) {
			mmah->n[i] = (i == ngpus-1) ? n-ceil(n/ngpus)*(ngpus-1) : ceil(n/ngpus);
			mmah->m[i] = m;
			cudaSetDevice (i);
			cudaMallocPitch ((void **) &mmah->devAddr[i], &pitch, mmah->word_size * mmah->m[i], mmah->n[i]);
			mmah->ld[i] = pitch/mmah->word_size;
		}
	}
}

void mublasSetVectorMgpu (mublasMgpuVectorHandle_t* mmah, void* hstAddr) {
	size_t i;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cublasSetVector (mmah->n[i], mmah->word_size, hstAddr, 1, mmah->devAddr[i], 1);
	}
}

void mublasSetVectorMgpuDistributed (mublasMgpuVectorHandle_t* mmah, void* hstAddr) {
	size_t i;
	char* mgpu_hstAddr = (char*)hstAddr;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cublasSetVector (mmah->n[i], mmah->word_size, mgpu_hstAddr, 1, mmah->devAddr[i], 1);
		mgpu_hstAddr += mmah->n[i] * mmah->word_size;
	}
}

void mublasSetMatrixMgpuDistributed (mublasMgpuMatrixHandle_t* mmah, char trans, void* hstAddr, size_t ld_hst) {
	size_t i;
	char* mgpu_hstAddr = (char*)hstAddr;
	if (trans == 'N' || trans == 'n') {
		for (i = 0; i < mmah->ngpus; i++) {
			cudaSetDevice (i);
			cublasSetMatrix (mmah->m[i], mmah->n[i], mmah->word_size, mgpu_hstAddr, ld_hst, mmah->devAddr[i], mmah->ld[i]);
			mgpu_hstAddr += mmah->m[i] * mmah->word_size;
		}
	} else {
		for (i = 0; i < mmah->ngpus; i++) {
			cudaSetDevice (i);
			cublasSetMatrix (mmah->m[i], mmah->n[i], mmah->word_size, mgpu_hstAddr, ld_hst, mmah->devAddr[i], mmah->ld[i]);
			mgpu_hstAddr += ld_hst * mmah->n[i] * mmah->word_size;
		}
	}
}

void mublasGetVectorMgpuDistributed (mublasMgpuVectorHandle_t* mmah, void* hstAddr) {
	size_t i;
	char* mgpu_hstAddr = (char*)hstAddr;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cublasGetVector (mmah->n[i], mmah->word_size, mmah->devAddr[i], 1, mgpu_hstAddr, 1);
		mgpu_hstAddr += mmah->n[i] * mmah->word_size;
	}
}

void mublasFreeMatrixMgpu (mublasMgpuMatrixHandle_t* mmah) {
	size_t i;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cudaFree (mmah->devAddr[i]);
	}
}

void mublasFreeVectorMgpu (mublasMgpuVectorHandle_t* mmah) {
	size_t i;
	for (i = 0; i < mmah->ngpus; i++) {
		cudaSetDevice (i);
		cudaFree (mmah->devAddr[i]);
	}
}

