/* ***********************************************
 
	GEMV for CUDA 

   *********************************************** */   

#include "mublas_devel.h"
#include "mublas_gemv_mgpu_n_kernel.cu"
#include "mublas_gemv_mgpu_t_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   

__host__ int32_t
MUBLAS_GEMV_MGPU_V2_SINGLE (
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 *alpha,
	const TYPE0 *a,
	const size_t lda,
	const TYPE0 *x,
	const size_t incx,
	const TYPE0 *beta,
	TYPE0 *y,
	const size_t incy
) {
	uint32_t kernel = 0;
	uint32_t use_vector = 0;
	dim3 threads;
	dim3 grid;
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (m == 0 || n == 0 || (*alpha == ZERO0 && *beta == ONE0 )) 
		return 0;
	if (*alpha == ZERO0) {
		MUBLAS_SCAL (mh, m, beta, y, incy);
		return 0;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration Phase 0
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if      (VEC4) mh->N_T = 4;
	else if (VEC2) mh->N_T = 2;
	if (
		(((long)a%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)x%(128/sizeof(TYPE0)))%16 == 0) && 
		(((long)y%(128/sizeof(TYPE0)))%16 == 0) &&
		(lda%mh->N_T == 0)
		) {
		use_vector = 1;
	} else {
		use_vector = 0;
	}
	mh->m = m;
	mh->n = n;
	// ---------------------------------------------

	// ---------------------------------------------
	kernel = 999; // default=CUBLAS;
	if (incx == 1 && incy == 1) {
		if (trans == 'N' || trans == 'n') {
			mh->BlkDim = 2;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_N;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_N;
			mh->func_getTxMax     = &getTxMax_GEMV_N;
			#include "mublas_gemv_n.incl"
			if (mh->N_T == 4) { 
				if (use_vector) kernel = 7140;
				else			kernel = 7141;
			} else if (mh->N_T == 2) { 
				if (use_vector) kernel = 7120;
				else 			kernel = 7121;
			} else {
								kernel = 7110;
			}
		} else if (trans == 'T' || trans == 't') {
			mh->BlkDim = 1;
			mh->func_getShmPerBlk = &getShmPerBlk_GEMV_T;
			mh->func_getBlkPerGrd = &getBlkPerGrd_GEMV_T;
			mh->func_getTxMax     = &getTxMax_GEMV_T;
			#include "mublas_gemv_t.incl"
			if (mh->N_T == 4) { 
				if (use_vector) kernel = 7240;
				else			kernel = 7210;
			} else if (mh->N_T == 2) {
				if (use_vector) kernel = 7220;
				else			kernel = 7210;
			} else {
								kernel = 7210;
			}
		}
	}
	// ---------------------------------------------

//	if (mh->kernelnumber == 999)  // force to invoke CUBLAS kernel
//		kernel = 999;
	
	uint32_t ngpus = mh->mgpu_ngpus;
	uint32_t gpuid = mh->mgpu_gpuid;
	size_t bs = mh->mgpu_bs;

	// ---------------------------------------------
	switch (kernel) {
		// N-kernel
		case 7140:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_N_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_N_DEVICE_v4, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_MGPU_V2_N_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		case 7141:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_N_DEVICE_s4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_N_DEVICE_s4, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_MGPU_V2_N_DEVICE_s4 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		case 7120:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_N_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_N_DEVICE_v2, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_MGPU_V2_N_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		case 7121:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_N_DEVICE_s2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_N_DEVICE_s2, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_MGPU_V2_N_DEVICE_s2 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		case 7110:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_N_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_N_DEVICE_s1, cudaFuncCachePreferShared);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->autotuning!=1) MUBLAS_GEMV_MGPU_V2_N_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		// T-kernel
		case 7240:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_T_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_T_DEVICE_v4, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->cc >= 300) MUBLAS_GEMV_MGPU_V2_T_DEVICE_v4 <<< grid, threads, 0, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			else MUBLAS_GEMV_MGPU_V2_T_DEVICE_v4 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		case 7220:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_T_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_T_DEVICE_v2, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->cc >= 300) MUBLAS_GEMV_MGPU_V2_T_DEVICE_v2 <<< grid, threads, 0, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			else MUBLAS_GEMV_MGPU_V2_T_DEVICE_v2 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		case 7210:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_GEMV_MGPU_V2_T_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_GEMV_MGPU_V2_T_DEVICE_s1, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			if (mh->cc >= 300) MUBLAS_GEMV_MGPU_V2_T_DEVICE_s1 <<< grid, threads, 0, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			else MUBLAS_GEMV_MGPU_V2_T_DEVICE_s1 <<< grid, threads, mh->ShmPerBlk, mh->streamId >>> (m, n, *alpha, a, lda, x, *beta, y, ngpus, gpuid, bs);
			break;
		// CUBLAS
		default:
			#if defined (CUBLAS_GEMV)
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetStream (ch, mh->streamId);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			cublasOperation_t tranAcublas = CUBLAS_OP_N;
			if (trans == 'T' || trans == 't') tranAcublas = CUBLAS_OP_T;
			if (trans == 'C' || trans == 'c') tranAcublas = CUBLAS_OP_C;
			CUBLAS_GEMV (ch, tranAcublas, m, n, alpha, a, lda, x, incx, beta, y, incy);
			#else
			printf ("mublas error: not implemented");
			#endif
			break;
	}
	// ---------------------------------------------

	mh->kernelnumber = kernel;

	return 0;
}

__host__ int32_t
MUBLAS_GEMV_MGPU_V2 ( // kblas compatible
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 alpha,
	TYPE0 **a,
	const size_t lda,
	TYPE0 **x,
	const size_t incx,
	const TYPE0 beta,
	TYPE0 **y,
	const size_t incy,
	const uint32_t ngpus,
	const size_t bs
) {
	uint32_t igpu;
	size_t nr, ni[NGPU_MAX] = {0};
	TYPE0 betai;

	nr = n;
	igpu = ngpus;
	while (nr >= bs) {
		ni[igpu%ngpus] += bs;
		nr -= bs;
		igpu++;
	} 
	ni[igpu%ngpus] += nr;

	if (trans == 'n' || trans == 'N') {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			betai = ZERO0;
			if (ni[igpu] == 0) {
				MUBLAS_SCAL (&mh[igpu], m, &betai, y[igpu], 1);
				break;
			}
			if (igpu == 0) betai = beta;
			MUBLAS_GEMV_MGPU_V2_SINGLE (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
		}
	} else {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			betai = ZERO0;
			if (igpu == 0) betai = beta;
			MUBLAS_SCAL (&mh[igpu], n, &betai, y[igpu], 1);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			if (igpu == 0) betai = ONE0;
			if (ni[igpu] != 0) MUBLAS_GEMV_MGPU_V2_SINGLE (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
		}
	}

	for(igpu = 0; igpu < ngpus; igpu++) {
		cudaSetDevice (igpu);
		cudaDeviceSynchronize();
	}
	return 0;
}

__host__ int32_t
MUBLAS_GEMV_MGPU_V2_ASYNC ( // kblas compatible
	mublasHandle_t* mh,	
	const char trans,
	const size_t m,
	const size_t n,
	const TYPE0 alpha,
	TYPE0 **a,
	const size_t lda,
	TYPE0 **x,
	const size_t incx,
	const TYPE0 beta,
	TYPE0 **y,
	const size_t incy,
	const uint32_t ngpus,
	const size_t bs
) {
	uint32_t igpu;
	size_t nr, ni[NGPU_MAX] = {0};
	TYPE0 betai;

	nr = n;
	igpu = ngpus;
	while (nr >= bs) {
		ni[igpu%ngpus] += bs;
		nr -= bs;
		igpu++;
	} 
	ni[igpu%ngpus] += nr;

	if (trans == 'n' || trans == 'N') {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			betai = ZERO0;
			if (ni[igpu] == 0) {
				MUBLAS_SCAL (&mh[igpu], m, &betai, y[igpu], 1);
			} else {
				if (igpu == 0) betai = beta;
				MUBLAS_GEMV_MGPU_V2_SINGLE (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
			}
		}
	} else {
		for (igpu = 0; igpu < ngpus; igpu++) {
			cudaSetDevice (igpu);
			betai = ZERO0;
			if (igpu == 0) betai = beta;
			MUBLAS_SCAL (&mh[igpu], n, &betai, y[igpu], 1);
			mh[igpu].mgpu_ngpus = ngpus;
			mh[igpu].mgpu_gpuid = igpu;
			mh[igpu].mgpu_bs = bs;
			if (igpu == 0) betai = ONE0;
			if (ni[igpu] != 0) MUBLAS_GEMV_MGPU_V2_SINGLE (&mh[igpu], trans, m, ni[igpu], &alpha, a[igpu], lda, x[igpu], incx, &betai, y[igpu], incy);
		}
	}
	return 0;

}

