/* ***********************************************

SCAL

memo:
Actually v2&v4 kernels are not needed.
v1 kernel with AT is enought to achieve best throughput.

  *********************************************** */   

#include "mublas_devel.h"
#include "mublas_scal_kernel.cu"

/* ***********************************************
	Host Interface
   *********************************************** */   
__host__ int32_t
MUBLAS_SCAL (
	mublasHandle_t* mh,	
	const size_t n,
	const TYPE0 *alpha,
	TYPE0 *x,
	const size_t incx
) {
	uint32_t kernel = 0;
	dim3 threads;
	dim3 grid;
	mublasRecipeDefault (mh);

	// ---------------------------------------------
	// 1. Shortcut
	if (n == 0) 
		return 0;
	// ---------------------------------------------

	// ---------------------------------------------
	// 2. Kernel Configuration Phase 0
	mh->N_one = sizeof (TYPE0);
	mh->N_T = 1;
	if (((long)x%(128/sizeof(TYPE0)))%16 == 0) {
		if (VEC4 && n%4 == 0)
			mh->N_T = 4;
		else if (VEC2 && n%2 == 0)
			mh->N_T = 2;
	}
	mh->n = n;
	// ---------------------------------------------

	// ---------------------------------------------
	// 3. Kernel Configuration Phase 1
	kernel = 999; // default=CUBLAS;
	if (incx == 1) {
		mh->BlkDim = 1;
		mh->func_getShmPerBlk = &getShmPerBlk_SCAL;
		mh->func_getBlkPerGrd = &getBlkPerGrd_SCAL;
		mh->func_getTxMax     = &getTxMax_SCAL;
		#include "mublas_scal.incl"
		if (mh->N_T == 4)
			kernel = 140;
		else if (mh->N_T == 2)
			kernel = 120;
		else
			kernel = 110;
	}
	// ---------------------------------------------

	// ---------------------------------------------
	// 4. Kernel Configuration Phase 2
	switch (kernel) {
		case 140:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_SCAL_DEVICE_v4);
			cudaFuncSetCacheConfig (MUBLAS_SCAL_DEVICE_v4, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			MUBLAS_SCAL_DEVICE_v4 <<< grid, threads, 0, mh->streamId >>> (n, *alpha, x);
			break;
		case 120:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_SCAL_DEVICE_v2);
			cudaFuncSetCacheConfig (MUBLAS_SCAL_DEVICE_v2, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			MUBLAS_SCAL_DEVICE_v2 <<< grid, threads, 0, mh->streamId >>> (n, *alpha, x);
			break;
		case 110:
			cudaFuncGetAttributes (&mh->funcatb, (void*)MUBLAS_SCAL_DEVICE_s1);
			cudaFuncSetCacheConfig (MUBLAS_SCAL_DEVICE_s1, cudaFuncCachePreferL1);
			mublas_optimizer (mh);
			threads = dim3 (mh->Tx, mh->Ty);
			grid = dim3 (mh->Bx, mh->By);
			MUBLAS_SCAL_DEVICE_s1 <<< grid, threads, 0, mh->streamId >>> (n, *alpha, x);
			break;
		default:
			#if defined (CUBLAS_SCAL)
			cublasHandle_t ch;
			cublasCreate (&ch);
			cublasSetStream (ch, mh->streamId);
			cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
			CUBLAS_SCAL (ch, n, alpha, x, incx);
			#else
			printf ("mublas error: not implemented");
			#endif
			break;
	}
	// ---------------------------------------------
		
	mh->kernelnumber = kernel;
	return 0;
}

