set term postscript eps enhanced color

set style line 11 lt 1 lw 1.0  ps 0.5 pt 6 lc rgb "#FF9966" # red
set style line 12 lt 2 lw 1.0  ps 0.5 pt 5 lc rgb "#FF6633" # red
set style line 13 lt 3 lw 1.0  ps 0.5 pt 4 lc rgb "#FF3300" # red
set style line 14 lt 4 lw 1.0  ps 0.5 pt 3 lc rgb "#CC0000" # red
set style line 21 lt 1 lw 1.0  ps 0.5 pt 2 lc rgb "#009900" # green
set style line 22 lt 2 lw 1.0  ps 0.5 pt 2 lc rgb "#09900" # green
set style line 31 lt 1 lw 1.0  ps 0.5 pt 8 lc rgb "#6699FF" # blue
set style line 32 lt 2 lw 1.0  ps 0.5 pt 7 lc rgb "#3366FF" # blue
set style line 33 lt 3 lw 1.0  ps 0.5 pt 9 lc rgb "#0033FF" # blue
set style line 34 lt 4 lw 1.0  ps 0.5 pt 1 lc rgb "#0000CC" # blue

##################################################################################################################
# Settings
##################################################################################################################

set nokey
set grid
#set size 2,2
set size 0.55,0.6
set xrange [0:10000]
set yrange [0:800]
set xlabel "Problem Size (M=N)" offset 0.5,0
#set ylabel "GFlops/s"
set ylabel "GB/s" offset 1.5,0

##################################################################################################################
# plot
##################################################################################################################

set title "SGEMV-N (Tesla K20Xm)" offset 0,-0.5
set output "fig_testing_mgpu_sgemv_n.eps"
plot \
"_mublas_sgemv_n_1gpu.dat" index 0 using 2:6 title "MUBLAS (1GPU)" w linespoints ls 11,\
"_mublas_sgemv_n_2gpu.dat" index 0 using 2:6 title "MUBLAS (2GPUs)" w linespoints ls 12,\
"_mublas_sgemv_n_3gpu.dat" index 0 using 2:6 title "MUBLAS (3GPUs)" w linespoints ls 13,\
"_mublas_sgemv_n_4gpu.dat" index 0 using 2:6 title "MUBLAS (4GPUs)" w linespoints ls 14,\
"_kblas_sgemv_n_1gpu.dat" index 0 using 2:6 title "KBLAS (1GPU)" w linespoints ls 31,\
"_kblas_sgemv_n_2gpu.dat" index 0 using 2:6 title "KBLAS (2GPUs)" w linespoints ls 32,\
"_kblas_sgemv_n_3gpu.dat" index 0 using 2:6 title "KBLAS (3GPUs)" w linespoints ls 33,\
"_kblas_sgemv_n_4gpu.dat" index 0 using 2:6 title "KBLAS (4GPUs)" w linespoints ls 34

set title "DGEMV-N (Tesla K20Xm)"
set output "fig_testing_mgpu_dgemv_n.eps"
plot \
"_mublas_dgemv_n_1gpu.dat" index 0 using 2:6 title "MUBLAS (1GPU)" w linespoints ls 11,\
"_mublas_dgemv_n_2gpu.dat" index 0 using 2:6 title "MUBLAS (2GPUs)" w linespoints ls 12,\
"_mublas_dgemv_n_3gpu.dat" index 0 using 2:6 title "MUBLAS (3GPUs)" w linespoints ls 13,\
"_mublas_dgemv_n_4gpu.dat" index 0 using 2:6 title "MUBLAS (4GPUs)" w linespoints ls 14,\
"_kblas_dgemv_n_1gpu.dat" index 0 using 2:6 title "KBLAS (1GPU)" w linespoints ls 31,\
"_kblas_dgemv_n_2gpu.dat" index 0 using 2:6 title "KBLAS (2GPUs)" w linespoints ls 32,\
"_kblas_dgemv_n_3gpu.dat" index 0 using 2:6 title "KBLAS (3GPUs)" w linespoints ls 33,\
"_kblas_dgemv_n_4gpu.dat" index 0 using 2:6 title "KBLAS (4GPUs)" w linespoints ls 34


set title "SGEMV-T (Tesla K20Xm)"
set output "fig_testing_mgpu_sgemv_t.eps"
plot \
"_mublas_sgemv_t_1gpu.dat" index 0 using 2:6 title "MUBLAS (1GPU)" w linespoints ls 11,\
"_mublas_sgemv_t_2gpu.dat" index 0 using 2:6 title "MUBLAS (2GPUs)" w linespoints ls 12,\
"_mublas_sgemv_t_3gpu.dat" index 0 using 2:6 title "MUBLAS (3GPUs)" w linespoints ls 13,\
"_mublas_sgemv_t_4gpu.dat" index 0 using 2:6 title "MUBLAS (4GPUs)" w linespoints ls 14,\
"_kblas_sgemv_t_1gpu.dat" index 0 using 2:6 title "KBLAS (1GPU)" w linespoints ls 31,\
"_kblas_sgemv_t_2gpu.dat" index 0 using 2:6 title "KBLAS (2GPUs)" w linespoints ls 32,\
"_kblas_sgemv_t_3gpu.dat" index 0 using 2:6 title "KBLAS (3GPUs)" w linespoints ls 33,\
"_kblas_sgemv_t_4gpu.dat" index 0 using 2:6 title "KBLAS (4GPUs)" w linespoints ls 34

set key below
set title "DGEMV-T (Tesla K20Xm)"
set output "fig_testing_mgpu_dgemv_t.eps"
plot \
"_mublas_dgemv_t_1gpu.dat" index 0 using 2:6 title "MUBLAS (1GPU)" w linespoints ls 11,\
"_mublas_dgemv_t_2gpu.dat" index 0 using 2:6 title "MUBLAS (2GPUs)" w linespoints ls 12,\
"_mublas_dgemv_t_3gpu.dat" index 0 using 2:6 title "MUBLAS (3GPUs)" w linespoints ls 13,\
"_mublas_dgemv_t_4gpu.dat" index 0 using 2:6 title "MUBLAS (4GPUs)" w linespoints ls 14,\
"_kblas_dgemv_t_1gpu.dat" index 0 using 2:6 title "KBLAS (1GPU)" w linespoints ls 31,\
"_kblas_dgemv_t_2gpu.dat" index 0 using 2:6 title "KBLAS (2GPUs)" w linespoints ls 32,\
"_kblas_dgemv_t_3gpu.dat" index 0 using 2:6 title "KBLAS (3GPUs)" w linespoints ls 33,\
"_kblas_dgemv_t_4gpu.dat" index 0 using 2:6 title "KBLAS (4GPUs)" w linespoints ls 34



