#include "testing_common.h"
#include "testing_level1b.h"
#include "testing_common.cpp"

int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	GPU_TYPE *hst_X, *hst_Y;
	GPU_TYPE *hst_Y_t;
	GPU_TYPE *dev_X, *dev_Y;
	double sec, t0, t1;
	cudaError_t err;

	size_t size_r = sizeof (REFERENCE_TYPE);
	REFERENCE_TYPE *hst_X_r = 0, *hst_Y_r = 0;
	GPU_TYPE result;
	REFERENCE_TYPE result_r;

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	dim_hst_setup (&th);
	print_info (&th);
// --------------------------------------------

// library setup ------------------------------
	#if defined (USE_NVML)
	uint32_t temp1 = 0, temp2 = 0;
	nvmlDevice_t dev;
	nvmlInit();
	nvmlDeviceGetHandleByIndex(th.devno, &dev);
	do {
		sleep(1);
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
	} while (temp1 >= TEMPLIMIT);
	#endif

	#if defined (CHECK_CUBLAS)
	cublasHandle_t ch;
	cublasCreate (&ch);
	cublasSetPointerMode(ch, CUBLAS_POINTER_MODE_HOST);
	#endif
	#if defined (MUBLAS)
	mublasHandle_t ha;
	mublasCreate (&ha);
	set_mublas_handler (&th, &ha);
	#elif defined (CUBLAS)
	cublasHandle_t ha;
	cublasCreate (&ha);
	cublasSetPointerMode(ha, CUBLAS_POINTER_MODE_HOST);
	if (th.cublasatomic == 1) 
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_ALLOWED);
	else
		cublasSetAtomicsMode(ha, CUBLAS_ATOMICS_NOT_ALLOWED);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
	cudaMallocHost ((void **) &hst_X, th.word_size * th.dim_n * th.incx);
	cudaMallocHost ((void **) &hst_Y, th.word_size * th.dim_n * th.incy);
	init_vector (th.dim_n * th.incx, hst_X);
	init_vector (th.dim_n * th.incy, hst_Y);
	if (th.mode == 'c') {
		cudaMallocHost ((void **) &hst_Y_t, th.word_size * th.dim_n * th.incy);
		hst_X_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_n * th.incx);
		hst_Y_r = (REFERENCE_TYPE *) malloc (size_r * th.dim_n * th.incy);
		copy_vector (th.dim_n * th.incx, hst_X, hst_X_r);
		copy_vector (th.dim_n * th.incy, hst_Y, hst_Y_r);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
	dim_setup (&th);
	while (1) {
		if (th.const_n == 0 && th.dim_n > th.dim_stop) break;

		cudaMalloc ((void **) &dev_X, th.word_size * th.dim_n * th.incx);
		cudaMalloc ((void **) &dev_Y, th.word_size * th.dim_n * th.incy);

		cublasSetVector(th.dim_n * th.incx, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector(th.dim_n * th.incy, th.word_size, hst_Y, 1, dev_Y, 1);

		#if defined (USE_NVML)
		nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp1);
		#endif

		printf ("--\t%zd\t--", (size_t)th.dim_n);
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = get_num_iter (&th);
			ROUTINE_TARGET(ha,th,th.dim_n,dev_X,dev_Y,result); // dummy
			// +++++ measurement +++++++
			t0 = gettime ();
			for (i = 0; i < iter; i++) ROUTINE_TARGET(ha,th,th.dim_n,dev_X,dev_Y,result);
			t1 = gettime ();
			// +++++++++++++++++++++++++
			sec = (t1 - t0) / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				t0 = gettime ();
				ROUTINE_TARGET(ha,th,th.dim_n,dev_X,dev_Y,result);
				t1 = gettime ();
				sec_all_exec[i] = (t1 - t0);
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec) sec = sec_all_exec[i];
			}
			#endif
			#if defined (USE_NVML)
			nvmlDeviceGetTemperature (dev, NVML_TEMPERATURE_GPU, &temp2);
			#endif
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
		}
		if (th.mode == 'c' || th.mode == 't') 
			ROUTINE_TARGET(ha,th,th.dim_n,dev_X,dev_Y,result);
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') 
			cublasGetVector (th.dim_m * th.incy, th.word_size, dev_Y, 1, hst_Y_t, 1);

		cudaFree (dev_X);
		cudaFree (dev_Y);

		if (th.mode == 'c') {
			#if (defined (CHECK_CUBLAS))
			REFERENCE_TYPE *dev_X_r, *dev_Y_r;
			copy_vector (th.dim_n * th.incy, hst_Y, hst_Y_r);
			cudaMalloc ((void **) &dev_X_r, size_r * th.dim_n * th.incx);
			cudaMalloc ((void **) &dev_Y_r, size_r * th.dim_n * th.incy);
			cublasSetVector (th.dim_n * th.incx, size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (th.dim_n * th.incy, size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(th,th.dim_n,hst_X_r,hst_Y_r,result_r);
			check_result (1, &result, &result_r);
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
			#else
			cublasGetVector (th.dim_n * th.incy, th.word_size, dev_Y, 1, hst_Y_t, 1);
			copy_vector (th.dim_n * th.incy, hst_Y, hst_Y_r);
			result_r = ROUTINE_REFERENCE(th,th.dim_n,hst_X_r,hst_Y_r);
			check_result (1, &result, &result_r);
			#endif
		}

		#if defined (MUBLAS)
		print_mublas_info1 (&ha);
		#endif
		#if defined (USE_NVML)
		printf ("\t%10u\t%10u", temp1, temp2);
		#endif
		printf ("\n");

		dim_increment (&th);
	}
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		free (hst_X_r);
		free (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (MUBLAS)
	print_mublas_info2 (&th, &ha);
	mublasDestroy (&ha);
	#elif defined (CUBLAS)
	cublasDestroy (ha);
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif
	#if defined (USE_NVML)
	nvmlShutdown();
	#endif
	cudaThreadExit();
// --------------------------------------------

	#endif
	return 0;
}
