#include <lis.h>
#include <cusparse_v2.h>

#include "testing_common.h"
#include "testing_level2s.h"
#include "testing_common.cpp"

#define TESTLIS

int32_t
main (int32_t argc, char **argv)
{
	size_t i, iter = MINLOOP;
	GPU_TYPE alpha, beta;
	GPU_TYPE *hst_A, *hst_X, *hst_Y;
	GPU_TYPE *hst_Y_t;
	GPU_TYPE *dev_A, *dev_X, *dev_Y;
	int32_t *hst_Aind, *hst_Aptr;
	int32_t *dev_Aind, *dev_Aptr;
	double sec, t0, t1;
	cudaError_t err;

	REFERENCE_TYPE *hst_A_r = 0, *hst_X_r = 0, *hst_Y_r = 0, alpha_r = REFERENCE_ZERO, beta_r = REFERENCE_ZERO;
	size_t size_r = sizeof (REFERENCE_TYPE);

// testing setup ------------------------------
	testingHandle_t th;
	testingCreate (argc, argv, &th);
	th.mode = 'p';
	th.nodisp = 1;

	print_info1 (&th);
// --------------------------------------------
//	th.TxStatic = 128;
//	th.TyStatic = 1;

// library setup ------------------------------
	#if defined (CHECK_CUBLAS)
	cusparseHandle_t ch;
	cusparseCreate (&ch);
	#endif
	#if defined (CUBLAS)
	cusparseHandle_t ha;
	cusparseCreate (&ha);
	#endif
	#if defined (CUBLAS) || defined (CHECK_CUBLAS)
	cusparseMatDescr_t descrA;
	cusparseCreateMatDescr (&descrA);
	cusparseSetMatType (descrA, CUSPARSE_MATRIX_TYPE_GENERAL);
	cusparseSetMatIndexBase (descrA, CUSPARSE_INDEX_BASE_ZERO);
	#endif
	#if defined (MUBLAS)
	mublasHandle_t ha;
	mublasCreate (&ha);
	set_mublas_handler (&th, &ha);
	#elif defined (CUBLAS)
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	magma_init();
	#elif defined (CULA)
    culaInitialize();
	#elif defined (ASPEN)
	ASPEN_init(0);
	#else
	#endif
#if DEBUG_ALLSKIP == 0
// --------------------------------------------

// memory setup -------------------------------
//	matrix setup
	printf ("%%");
#ifdef TESTLIS
	LIS_MATRIX lis_A;
	lis_initialize (&argc, &argv);
	lis_matrix_create (0, &lis_A);
	lis_matrix_set_type (lis_A, LIS_MATRIX_CSR);
	lis_input_matrix (lis_A, argv[1]);
	lis_matrix_assemble(lis_A);
	th.dim_k_dev = lis_A->nnz;
	th.dim_n_dev = lis_A->n;
	th.dim_m_dev = th.dim_n_dev;//lis_A->m;
#endif
// end
	size_t size_n	= sizeof (GPU_TYPE) * th.dim_n_dev;
	size_t size_m	= sizeof (GPU_TYPE) * th.dim_m_dev;
	size_t size_mi	= sizeof (int32_t) * (th.dim_m_dev+1);
	size_t size_nnz	= sizeof (GPU_TYPE) * th.dim_k_dev;
	size_t size_nnzi= sizeof (int32_t) * th.dim_k_dev;
	cudaMallocHost ((void **) &hst_A, size_nnz);
	cudaMallocHost ((void **) &hst_Aptr, size_mi);
	cudaMallocHost ((void **) &hst_Aind, size_nnzi);
	cudaMallocHost ((void **) &hst_X, size_n);
	cudaMallocHost ((void **) &hst_Y, size_m);
//	mublasInitCSRMatrix (&th, hst_A, hst_Aptr, hst_Aind); // th.mtx_file, th.dim_k_dev(nnz), th.dim_m_dev, th.dim_n_dev
	mublasInit2D (th.dim_n_dev, 1, 0, hst_X);
	mublasInit2D (th.dim_m_dev, 1, 0, hst_Y);
	mublasInit2D (1, 1, 0, &alpha);
	mublasInit2D (1, 1, 0, &beta);
#ifdef TESTLIS
	for (i = 0; i < th.dim_k_dev; i++) hst_A[i] = lis_A->value[i];
	for (i = 0; i < th.dim_m_dev+1; i++) hst_Aptr[i] = lis_A->ptr[i];
	for (i = 0; i < th.dim_k_dev; i++) hst_Aind[i] = lis_A->index[i];
#endif
	if (th.mode == 'c') {
		cudaMallocHost ((void **) &hst_Y_t, size_m);
		hst_A_r = (REFERENCE_TYPE *) malloc (size_nnz);
		hst_X_r = (REFERENCE_TYPE *) malloc (size_n);
		hst_Y_r = (REFERENCE_TYPE *) malloc (size_m);
		mublasConvert2D (th.dim_k_dev, 1, 0, hst_A, hst_A_r);
		mublasConvert2D (th.dim_n_dev, 1, 0, hst_X, hst_X_r);
		mublasConvert2D (th.dim_m_dev, 1, 0, hst_Y, hst_Y_r);
		mublasConvert2D (1, 1, 0, &alpha, &alpha_r);
		mublasConvert2D (1, 1, 0, &beta, &beta_r);
	}
// --------------------------------------------

	print_info2 (&th);

// evaluation ---------------------------------
		cudaMalloc ((void **) &dev_A, size_nnz);
		cudaMalloc ((void **) &dev_Aptr, size_mi);
		cudaMalloc ((void **) &dev_Aind, size_nnzi);
		cudaMalloc ((void **) &dev_X, size_n);
		cudaMalloc ((void **) &dev_Y, size_m);

		cublasSetVector (th.dim_k_dev, th.word_size, hst_A, 1, dev_A, 1);
		cublasSetVector (th.dim_m_dev+1, sizeof (int32_t), hst_Aptr, 1, dev_Aptr, 1);
		cublasSetVector (th.dim_k_dev, sizeof (int32_t), hst_Aind, 1, dev_Aind, 1);
		cublasSetVector (th.dim_n_dev, th.word_size, hst_X, 1, dev_X, 1);
		cublasSetVector (th.dim_m_dev, th.word_size, hst_Y, 1, dev_Y, 1);

		printf ("%zd\t%zd\t--", (size_t)th.dim_m_dev, (size_t)th.dim_n_dev);
		get_routine_theoretial_performance (&th);

		// execution ---------------------------------
		if (th.mode == 'p') {
			#if PERF_ACCURATE
			iter = 100;//get_num_iter (&th);
			ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
			// +++++ measurement +++++++
			t0 = gettime ();
			for (i = 0; i < iter; i++) ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
			t1 = gettime ();
			// +++++++++++++++++++++++++
			sec = (t1 - t0) / (float)iter;
			#else
			float sec_all_exec[MINLOOP];
			iter = MINLOOP;
			for (i = 0; i < iter; i++) {
				t0 = gettime ();
				ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
				t1 = gettime ();
				sec_all_exec[i] = (t1 - t0);
			}
			sec = 0.;
			for (i = 0; i < iter; i++) {
				if (sec_all_exec[i] > sec)
					sec = sec_all_exec[i];
			}
			#endif
			printf ("\t%1.3e\t%1.3e\t%1.3e\t(%zd)", sec, (th.routine_flops / sec) * 1.0e-9, (th.routine_bytes / sec) * 1.0e-9, (size_t)iter);
		}
		if (th.mode == 'c' || th.mode == 't') 
			ROUTINE_TARGET(ha,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha,descrA,dev_A,dev_Aptr,dev_Aind,dev_X,beta,dev_Y);
		err = cudaGetLastError();
		if (err != cudaSuccess) printf("error : %s\n", cudaGetErrorString(err));
		// -------------------------------------------

		if (th.mode == 'c') 
			cublasGetVector (th.dim_m_dev, th.word_size, dev_Y, 1, hst_Y_t, 1);

		cudaFree (dev_A);
		cudaFree (dev_Aptr);
		cudaFree (dev_Aind);
		cudaFree (dev_X);
		cudaFree (dev_Y);

		if (th.mode == 'c') {
			REFERENCE_TYPE *dev_A_r, *dev_X_r, *dev_Y_r;
			int32_t *dev_Aptr_r, *dev_Aind_r;
			mublasConvert2D (th.dim_m_dev, 1, 0, hst_Y, hst_Y_r);
			cudaMalloc ((void **) &dev_A_r, size_nnz);
			cudaMalloc ((void **) &dev_Aptr_r, size_mi);
			cudaMalloc ((void **) &dev_Aind_r, size_nnzi);
			cudaMalloc ((void **) &dev_X_r, size_n);
			cudaMalloc ((void **) &dev_Y_r, size_m);
			cublasSetVector (th.dim_k_dev, size_r, hst_A_r, 1, dev_A_r, 1);
			cublasSetVector (th.dim_m_dev+1, sizeof (int32_t), hst_Aptr, 1, dev_Aptr_r, 1);
			cublasSetVector (th.dim_k_dev, sizeof (int32_t), hst_Aind, 1, dev_Aind_r, 1);
			cublasSetVector (th.dim_n_dev, size_r, hst_X_r, 1, dev_X_r, 1);
			cublasSetVector (th.dim_m_dev, size_r, hst_Y_r, 1, dev_Y_r, 1);
			ROUTINE_REFERENCE(ch,th,th.dim_m_dev,th.dim_n_dev,th.dim_k_dev,alpha_r,descrA,dev_A_r,dev_Aptr_r,dev_Aind_r,dev_X_r,beta_r,dev_Y_r);
			cublasGetVector (th.dim_m_dev, size_r, dev_Y_r, 1, hst_Y_r, 1);
			mublasCheckVector (th.dim_m_dev, hst_Y_t, hst_Y_r);
			cudaFree (dev_A_r); 
			cudaFree (dev_Aptr_r); 
			cudaFree (dev_Aind_r); 
			cudaFree (dev_X_r);
			cudaFree (dev_Y_r);
		}

		#if defined (MUBLAS)
		print_mublas_info1 (&th, &ha);
		#endif
#if DISPINFO
		printf ("\n");
#else
#ifdef MUBLAS
		printf ("\t%zd\t", ha.Tx);
#else
		printf ("\t");
#endif
#endif
// --------------------------------------------

// shutdown -----------------------------------
	cudaFreeHost (hst_A);
	cudaFreeHost (hst_Aptr);
	cudaFreeHost (hst_Aind);
	cudaFreeHost (hst_X);
	cudaFreeHost (hst_Y);
	if (th.mode == 'c') {
		free (hst_A_r);
		free (hst_X_r);
		free (hst_Y_r);
		cudaFreeHost (hst_Y_t);
	}

	#if defined (CHECK_CUBLAS)
	cusparseDestroy (ch);
	#endif
	#if defined (CUBLAS) 
	cusparseDestroy (ha);
	#endif
	#if defined (MUBLAS)
	print_mublas_info2 (&th, &ha);
	mublasDestroy (&ha);
	#elif defined (CUBLAS)
	#elif defined (KBLAS)
	#elif defined (MAGMA)
	#elif defined (CULA)
	#elif defined (ASPEN)
	ASPEN_shutdown();
	#endif

#ifdef TESTLIS
	lis_matrix_destroy (lis_A);
	lis_finalize ();
#endif
// --------------------------------------------

#endif
	return 0;
}

